/*
** Compilieren mit:
**    gcc -Wall -o hello -lpthread hello.c
*/

#include <iostream>
#include <pthread.h>
#include <stdlib.h>
#include <sys/time.h>

using namespace std;

#define N 8000
#define THREADS 4

double a[N][N];
volatile double s[THREADS];

double getTime()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec + tv.tv_usec * 0.000001;
}

void * WorkS(void *arg)
{
	int p = (int)(long)arg;
	int i, j;

	for (i=p*N/THREADS; i<(p+1)*N/THREADS; i++) {
		for (j=0; j<N; j++) {
			s[p] += a[i][j];
		}
	}
	return NULL;
}

void * WorkL(void *arg)
{
	int p = (int)(long)arg;
	int i, j;
	volatile double sl = 0;

	for (i=p*N/THREADS; i<(p+1)*N/THREADS; i++) {
		for (j=0; j<N; j++) {
			sl += a[i][j];
		}
	}
	s[p] = sl;
	return NULL;
}

void WorkSeq()
{
	int i, j;

	for (i=0; i<N; i++) {
		for (j=0; j<N; j++) {
			s[0] += a[i][j];
		}
	}
}

int main(int argc, char **argv)
{
	pthread_t t[THREADS];
	int i,j;
	double t0;
	double sum;
	
	for (i=0; i<N; i++) {
		for (j=0; j<N; j++) {
			a[i][j] = i/(j+1);
		}
	}
	for (i=0; i<THREADS; i++)
		s[i] = 0;
	
	WorkSeq();
	s[0] = 0;
	t0 = getTime();
	WorkSeq();
	cout << "Zeit(seq): " << (getTime()-t0) << "s, sum = " << s[0] << "\n";
	s[0] = 0;

	sum = 0;
	t0 = getTime();
	for (i=0; i<THREADS; i++) {
		if (pthread_create(&t[i], NULL, WorkS, (void *)(long)i) != 0) {
			cerr << "Can't create thread\n";
			return 1;
		}
	}

	for (i=0; i<THREADS; i++) {
		pthread_join(t[i], NULL);
		sum += s[i];
	}

	cout << "Zeit(shared): " << (getTime()-t0) << "s, sum = " << s[0] << "\n";

	for (i=0; i<THREADS; i++)
		s[i] = 0;
	
	sum = 0;
	t0 = getTime();
	for (i=0; i<THREADS; i++) {
		if (pthread_create(&t[i], NULL, WorkL, (void *)(long)i) != 0) {
			cerr << "Can't create thread\n";
			return 1;
		}
	}

	for (i=0; i<THREADS; i++) {
		pthread_join(t[i], NULL);
		sum += s[i];
	}

	cout << "Zeit(lokal): " << (getTime()-t0) << "s, sum = " << s[0] << "\n";

	return 0;
}
